# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of the Deep Average-network Fictitious Play.

Coupled with agents that compute a best-response (BR) at each iteration, instead
of keeping in memory all the BRs from past iterations Deep Average-network
Fictitious Play learns along the way the policy generating the average
distribution. This is done by keeping a buffer of state-action pairs generated
by past BRs and learning the average policy (represented by a neural network) by
minimizing a categorical loss. This approach is inspired by the Neural
Fictitious Self Play (NFSP) method (Heinrich & Silver, 2016), developed
initially for imperfect information games with a finite number of players, and
adapted here to the MFG setting.
"""

import dataclasses
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import optax

from open_spiel.python import rl_agent
from open_spiel.python import rl_agent_policy
from open_spiel.python import rl_environment
from open_spiel.python.mfg.algorithms import distribution
import pyspiel
from open_spiel.python.utils import reservoir_buffer
from open_spiel.python.utils import training


@dataclasses.dataclass
class Transition:
  """Transitions stored in the reservoir buffer."""
  info_state: np.ndarray
  action_probs: np.ndarray
  legal_actions_mask: np.ndarray


class AveragePolicy(rl_agent.AbstractAgent):
  """NFSP-like agent that learns an average policy using a single network."""

  def __init__(self,
               player_id: int,
               br_rl_agent: rl_agent.AbstractAgent,
               state_representation_size: int,
               num_actions: int,
               hidden_layers_sizes: List[int],
               params_avg_network: Optional[jnp.ndarray] = None,
               reservoir_buffer_capacity: int = 100000,
               batch_size: int = 128,
               learning_rate: float = 0.01,
               min_buffer_size_to_learn: int = 1000,
               optimizer_str: str = 'sgd',
               gradient_clipping: Optional[float] = None,
               seed: int = 42,
               tau: float = 1.0):
    """Initialize the AveragePolicy agent."""
    self._br_rl_agent = br_rl_agent
    self._player_id = player_id
    self._num_actions = num_actions
    self._batch_size = batch_size
    self._min_buffer_size_to_learn = min_buffer_size_to_learn

    self._reservoir_buffer = reservoir_buffer.ReservoirBuffer(
        reservoir_buffer_capacity)

    # Keep track of the last training loss achieved in an update step.
    self._last_loss_value = None

    # Average policy network.
    def network(x):
      mlp = hk.nets.MLP(hidden_layers_sizes + [num_actions])
      return mlp(x)

    self.avg_network = hk.without_apply_rng(hk.transform(network))

    def avg_network_policy(param, info_state):
      action_values = self.avg_network.apply(param, info_state)
      return jax.nn.softmax(action_values / tau, axis=1)

    self._avg_network_policy = jax.jit(avg_network_policy)

    rng = jax.random.PRNGKey(seed)
    x = jnp.ones([1, state_representation_size])
    # Use the specified parameters if any, or initialize the network with random
    # weights.
    if params_avg_network is None:
      self._params_avg_network = self.avg_network.init(rng, x)
    else:
      self._params_avg_network = jax.tree_map(lambda x: x.copy(),
                                                   params_avg_network)
    self._params_avg_network = jax.device_put(self._params_avg_network)

    if optimizer_str == 'adam':
      optimizer = optax.adam(learning_rate)
    elif optimizer_str == 'sgd':
      optimizer = optax.sgd(learning_rate)
    else:
      raise ValueError('Not implemented, choose from "adam" and "sgd".')

    if gradient_clipping:
      optimizer = optax.chain(optimizer,
                              optax.clip_by_global_norm(gradient_clipping))

    opt_init, opt_update = optimizer.init, optimizer.update

    def opt_update_fn(params, opt_state, gradient):
      """Learning rule (stochastic gradient descent)."""
      updates, opt_state = opt_update(gradient, opt_state)
      new_params = optax.apply_updates(params, updates)
      return new_params, opt_state

    self._opt_update_fn = opt_update_fn
    self._opt_state = opt_init(self._params_avg_network)
    self._loss_and_grad = jax.value_and_grad(self._loss_avg, has_aux=False)

    self._jit_update = jax.jit(self._get_update_fn())

  def _get_update_fn(self):
    """Returns the function that updates the parameters."""

    def update(param_avg, opt_state_avg, info_states, action_probs):
      loss_val, grad_val = self._loss_and_grad(param_avg, info_states,
                                               action_probs)
      new_param_avg, new_opt_state_avg = self._opt_update_fn(
          param_avg, opt_state_avg, grad_val)
      return new_param_avg, new_opt_state_avg, loss_val

    return update

  def _act(self, info_state, legal_actions) -> Tuple[int, np.ndarray]:
    """Returns an action and the action probabilities."""
    info_state = np.reshape(info_state, [1, -1])
    action_probs = self._avg_network_policy(self._params_avg_network,
                                            info_state)
    # Remove illegal actions and normalize probs
    probs = np.zeros(self._num_actions)
    action_probs = np.asarray(action_probs)
    probs[legal_actions] = action_probs[0][legal_actions]
    probs /= sum(probs)
    action = np.random.choice(len(probs), p=probs)
    return action, probs

  @property
  def loss(self) -> Optional[float]:
    """Return the latest loss."""
    return self._last_loss_value

  def step(self,
           time_step: rl_environment.TimeStep,
           is_evaluation: bool = True) -> Optional[rl_agent.StepOutput]:
    """Returns the action to be taken by following the average network policy.

    Note that unlike most other algorithms, this method doesn't train the agent.
    Instead, we add new samples to the reservoir buffer and the training happens
    at a later stage.

    Args:
      time_step: an instance of rl_environment.TimeStep.
      is_evaluation: bool, whether this is a training or evaluation call.

    Returns:
      A `rl_agent.StepOutput` containing the action probs and chosen action.
    """

    # Prepare for the next episode.
    if time_step.last():
      return

    if is_evaluation:
      # Use the average policy network.
      info_state = time_step.observations['info_state'][self._player_id]
      legal_actions = time_step.observations['legal_actions'][self._player_id]
      action, probs = self._act(info_state, legal_actions)
      return rl_agent.StepOutput(action=action, probs=probs)

    # Use the best response agent and add the transition in the reservoir
    # buffer.
    br_agent_output = self._br_rl_agent.step(time_step, is_evaluation=True)
    self._add_transition(time_step, br_agent_output)
    return br_agent_output

  def _add_transition(self, time_step, agent_output):
    """Adds the new transition using `time_step` to the reservoir buffer.

    Transitions are in the form (time_step, agent_output.probs, legal_mask).

    Args:
      time_step: an instance of rl_environment.TimeStep.
      agent_output: an instance of rl_agent.StepOutput.
    """
    legal_actions = time_step.observations['legal_actions'][self._player_id]
    legal_actions_mask = np.zeros(self._num_actions)
    legal_actions_mask[legal_actions] = 1.0
    transition = Transition(
        info_state=(time_step.observations['info_state'][self._player_id][:]),
        action_probs=agent_output.probs,
        legal_actions_mask=legal_actions_mask)
    self._reservoir_buffer.add(transition)

  def _loss_avg(self, param_avg, info_states, action_probs):
    avg_logit = self.avg_network.apply(param_avg, info_states)
    loss_value = -jnp.sum(
        action_probs * jax.nn.log_softmax(avg_logit)) / avg_logit.shape[0]
    return loss_value

  def learn(self) -> Optional[float]:
    """Compute the loss on sampled transitions and perform a avg-network update.

    If there are not enough elements in the buffer, no loss is computed and
    `None` is returned instead.

    Returns:
      The average loss obtained on this batch of transitions or `None`.
    """
    if (len(self._reservoir_buffer) < self._batch_size or
        len(self._reservoir_buffer) < self._min_buffer_size_to_learn):
      return None

    transitions = self._reservoir_buffer.sample(self._batch_size)
    info_states = np.asarray([t.info_state for t in transitions])
    action_probs = np.asarray([t.action_probs for t in transitions])

    self._params_avg_network, self._opt_state, loss_val_avg = self._jit_update(
        self._params_avg_network, self._opt_state, info_states, action_probs)
    self._last_loss_value = float(loss_val_avg)
    return loss_val_avg


class AverageNetworkFictitiousPlay(object):
  """Deep Average-network Fictitious Play.

  See the file description for more information.
  """

  def __init__(self,
               game: pyspiel.Game,
               envs: Sequence[rl_environment.Environment],
               br_rl_agents: Sequence[rl_agent.AbstractAgent],
               num_episodes_per_iteration: int,
               num_training_steps_per_iteration: int,
               eval_every: int = 200,
               logging_fn: Optional[Callable[[int, int, Dict[str, Any]],
                                             None]] = None,
               **kwargs):
    """Initializes the greedy policy.

    Args:
      game: The game to analyze.
      envs: RL environment for each player.
      br_rl_agents: Best response, e.g. DQN, agents for each player.
      num_episodes_per_iteration: Number of episodes to collect samples that are
        added to the reservoir buffer.
      num_training_steps_per_iteration: Number of steps to train the average
        policy in each iteration.
      eval_every: Number of training steps between two evaluations.
      logging_fn: Callable for logging the metrics. The arguments will be the
        current iteration, episode and a dictionary of metrics to log.
      **kwargs: kwargs passed to the AveragePolicy() constructor.
    """
    self._game = game
    self._envs = envs
    self._num_episodes_per_iteration = num_episodes_per_iteration
    self._num_training_steps_per_iteration = num_training_steps_per_iteration
    self._eval_every = eval_every
    self._logging_fn = logging_fn

    self._num_players = game.num_players()
    self._fp_iteration = 0

    env = self._envs[0]
    info_state_size = env.observation_spec()['info_state'][0]
    num_actions = env.action_spec()['num_actions']

    self._avg_rl_agents = [
        AveragePolicy(p, br_rl_agents[p], info_state_size, num_actions,
                      **kwargs) for p in range(self._num_players)
    ]
    self._policy = rl_agent_policy.JointRLAgentPolicy(
        self._game,
        {idx: agent for idx, agent in enumerate(self._avg_rl_agents)},
        use_observation=env.use_observation)
    self._update_distribution()

  def _update_distribution(self):
    """Calculates the current distribution and updates the environments."""
    self._distribution = distribution.DistributionPolicy(
        self._game, self._policy)
    for env in self._envs:
      env.update_mfg_distribution(self._distribution)

  @property
  def policy(self) -> rl_agent_policy.JointRLAgentPolicy:
    return self._policy

  def iteration(self):
    """An average-network fictitious play step."""
    # Generate samples using latest best-response and add them to the reservoir
    # buffer. Note that the algorithm is agnostic to the best-response policies
    # as we only use them to collect new samples. They can be approximate (e.g.
    # backed by a deep algorithm) or exact.
    training.run_episodes(
        self._envs,
        self._avg_rl_agents,
        num_episodes=self._num_episodes_per_iteration,
        is_evaluation=False)

    # Train the average policy.
    for step in range(self._num_training_steps_per_iteration):
      for avg_rl_agent in self._avg_rl_agents:
        avg_rl_agent.learn()

      if self._logging_fn and (step + 1) % self._eval_every == 0:
        self._logging_fn(
            self._fp_iteration, step, {
                f'avg_agent{i}/loss': float(agent.loss)
                for i, agent in enumerate(self._avg_rl_agents)
            })

    # Update the distribution.
    self._update_distribution()
    self._fp_iteration += 1



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""An implementation of Information Set Monte Carlo Tree Search (IS-MCTS).

See Cowling, Powley, and Whitehouse 2011.
https://ieeexplore.ieee.org/document/6203567
"""

import copy
import enum
import numpy as np
import pyspiel

UNLIMITED_NUM_WORLD_SAMPLES = -1
UNEXPANDED_VISIT_COUNT = -1
TIE_TOLERANCE = 1e-5


class ISMCTSFinalPolicyType(enum.Enum):
  """A enumeration class for final ISMCTS policy type."""
  NORMALIZED_VISITED_COUNT = 1
  MAX_VISIT_COUNT = 2
  MAX_VALUE = 3


class ChildSelectionPolicy(enum.Enum):
  """A enumeration class for children selection in ISMCTS."""
  UCT = 1
  PUCT = 2


class ChildInfo(object):
  """Child node information for the search tree."""

  def __init__(self, visits, return_sum, prior):
    self.visits = visits
    self.return_sum = return_sum
    self.prior = prior

  def value(self):
    return self.return_sum / self.visits


class ISMCTSNode(object):
  """Node data structure for the search tree."""

  def __init__(self):
    self.child_info = {}
    self.total_visits = 0
    self.prior_map = {}


class ISMCTSBot(pyspiel.Bot):
  """Adapted from the C++ implementation."""

  def __init__(self,
               game,
               evaluator,
               uct_c,
               max_simulations,
               max_world_samples=UNLIMITED_NUM_WORLD_SAMPLES,
               random_state=None,
               final_policy_type=ISMCTSFinalPolicyType.MAX_VISIT_COUNT,
               use_observation_string=False,
               allow_inconsistent_action_sets=False,
               child_selection_policy=ChildSelectionPolicy.PUCT):

    pyspiel.Bot.__init__(self)
    self._game = game
    self._evaluator = evaluator
    self._uct_c = uct_c
    self._max_simulations = max_simulations
    self._max_world_samples = max_world_samples
    self._final_policy_type = final_policy_type
    self._use_observation_string = use_observation_string
    self._allow_inconsistent_action_sets = allow_inconsistent_action_sets
    self._nodes = {}
    self._node_pool = []
    self._root_samples = []
    self._random_state = random_state or np.random.RandomState()
    self._child_selection_policy = child_selection_policy
    self._resampler_cb = None

  def random_number(self):
    return self._random_state.uniform()

  def reset(self):
    self._nodes = {}
    self._node_pool = []
    self._root_samples = []

  def get_state_key(self, state):
    if self._use_observation_string:
      return state.current_player(), state.observation_string()
    else:
      return state.current_player(), state.information_state_string()

  def run_search(self, state):
    self.reset()
    assert state.get_game().get_type(
    ).dynamics == pyspiel.GameType.Dynamics.SEQUENTIAL
    assert state.get_game().get_type(
    ).information == pyspiel.GameType.Information.IMPERFECT_INFORMATION

    legal_actions = state.legal_actions()
    if len(legal_actions) == 1:
      return [(legal_actions[0], 1.0)]

    self._root_node = self.create_new_node(state)

    assert self._root_node

    root_infostate_key = self.get_state_key(state)

    for _ in range(self._max_simulations):
      # how to sample a pyspiel.state from another pyspiel.state?
      sampled_root_state = self.sample_root_state(state)
      assert root_infostate_key == self.get_state_key(sampled_root_state)
      assert sampled_root_state
      self.run_simulation(sampled_root_state)

    if self._allow_inconsistent_action_sets:  # when this happens?
      legal_actions = state.legal_actions()
      temp_node = self.filter_illegals(self._root_node, legal_actions)
      assert temp_node.total_visits > 0
      return self.get_final_policy(state, temp_node)
    else:
      return self.get_final_policy(state, self._root_node)

  def step(self, state):
    action_list, prob_list = zip(*self.run_search(state))
    return self._random_state.choice(action_list, p=prob_list)

  def get_policy(self, state):
    return self.run_search(state)

  def step_with_policy(self, state):
    policy = self.get_policy(state)
    action_list, prob_list = zip(*policy)
    sampled_action = self._random_state.choice(action_list, p=prob_list)
    return policy, sampled_action

  def get_final_policy(self, state, node):
    assert node
    if self._final_policy_type == ISMCTSFinalPolicyType.NORMALIZED_VISITED_COUNT:
      assert node.total_visits > 0
      total_visits = node.total_visits
      policy = [(action, child.visits / total_visits)
                for action, child in node.child_info.items()]
    elif self._final_policy_type == ISMCTSFinalPolicyType.MAX_VISIT_COUNT:
      assert node.total_visits > 0
      max_visits = -float('inf')
      count = 0
      for action, child in node.child_info.items():
        if child.visits == max_visits:
          count += 1
        elif child.visits > max_visits:
          max_visits = child.visits
          count = 1
      policy = [(action, 1. / count if child.visits == max_visits else 0.0)
                for action, child in node.child_info.items()]
    elif self._final_policy_type == ISMCTSFinalPolicyType.MAX_VALUE:
      assert node.total_visits > 0
      max_value = -float('inf')
      count = 0
      for action, child in node.child_info.items():
        if child.value() == max_value:
          count += 1
        elif child.value() > max_value:
          max_value = child.value()
          count = 1
      policy = [(action, 1. / count if child.value() == max_value else 0.0)
                for action, child in node.child_info.items()]

    policy_size = len(policy)
    legal_actions = state.legal_actions()
    if policy_size < len(legal_actions):  # do we really need this step?
      for action in legal_actions:
        if action not in node.child_info:
          policy.append((action, 0.0))
    return policy

  def sample_root_state(self, state):
    if self._max_world_samples == UNLIMITED_NUM_WORLD_SAMPLES:
      return self.resample_from_infostate(state)
    elif len(self._root_samples) < self._max_world_samples:
      self._root_samples.append(self.resample_from_infostate(state))
      return self._root_samples[-1].clone()
    elif len(self._root_samples) == self._max_world_samples:
      idx = self._random_state.randint(len(self._root_samples))
      return self._root_samples[idx].clone()
    else:
      raise pyspiel.SpielError(
          'Case not handled (badly set max_world_samples..?)')

  def resample_from_infostate(self, state):
    if self._resampler_cb:
      return self._resampler_cb(state, state.current_player())
    else:
      return state.resample_from_infostate(
          state.current_player(), pyspiel.UniformProbabilitySampler(0., 1.))

  def create_new_node(self, state):
    infostate_key = self.get_state_key(state)
    self._node_pool.append(ISMCTSNode())
    node = self._node_pool[-1]
    self._nodes[infostate_key] = node
    node.total_visits = UNEXPANDED_VISIT_COUNT
    return node

  def set_resampler(self, cb):
    self._resampler_cb = cb

  def lookup_node(self, state):
    if self.get_state_key(state) in self._nodes:
      return self._nodes[self.get_state_key(state)]
    return None

  def lookup_or_create_node(self, state):
    node = self.lookup_node(state)
    if node:
      return node
    return self.create_new_node(state)

  def filter_illegals(self, node, legal_actions):
    new_node = copy.deepcopy(node)
    for action, child in node.child_info.items():
      if action not in legal_actions:
        new_node.total_visits -= child.visits
        del new_node.child_info[action]
    return new_node

  def expand_if_necessary(self, node, action):
    if action not in node.child_info:
      node.child_info[action] = ChildInfo(0.0, 0.0, node.prior_map[action])

  def select_action_tree_policy(self, node, legal_actions):
    if self._allow_inconsistent_action_sets:
      temp_node = self.filter_illegals(node, legal_actions)
      if temp_node.total_visits == 0:
        action = legal_actions[self._random_state.randint(
            len(legal_actions))]  # prior?
        self.expand_if_necessary(node, action)
        return action
      else:
        return self.select_action(temp_node)
    else:
      return self.select_action(node)

  def select_action(self, node):
    candidates = []
    max_value = -float('inf')
    for action, child in node.child_info.items():
      assert child.visits > 0

      action_value = child.value()
      if self._child_selection_policy == ChildSelectionPolicy.UCT:
        action_value += (self._uct_c *
                         np.sqrt(np.log(node.total_visits)/child.visits))
      elif self._child_selection_policy == ChildSelectionPolicy.PUCT:
        action_value += (self._uct_c * child.prior *
                         np.sqrt(node.total_visits)/(1 + child.visits))
      else:
        raise pyspiel.SpielError('Child selection policy unrecognized.')
      if action_value > max_value + TIE_TOLERANCE:
        candidates = [action]
        max_value = action_value
      elif (action_value > max_value - TIE_TOLERANCE and
            action_value < max_value + TIE_TOLERANCE):
        candidates.append(action)
        max_value = action_value

    assert len(candidates) >= 1
    return candidates[self._random_state.randint(len(candidates))]

  def check_expand(self, node, legal_actions):
    if not self._allow_inconsistent_action_sets and len(
        node.child_info) == len(legal_actions):
      return pyspiel.INVALID_ACTION
    legal_actions_copy = copy.deepcopy(legal_actions)
    self._random_state.shuffle(legal_actions_copy)
    for action in legal_actions_copy:
      if action not in node.child_info:
        return action
    return pyspiel.INVALID_ACTION

  def run_simulation(self, state):
    if state.is_terminal():
      return state.returns()
    elif state.is_chance_node():
      action_list, prob_list = zip(*state.chance_outcomes())
      chance_action = self._random_state.choice(action_list, p=prob_list)
      state.apply_action(chance_action)
      return self.run_simulation(state)
    legal_actions = state.legal_actions()
    cur_player = state.current_player()
    node = self.lookup_or_create_node(state)

    assert node

    if node.total_visits == UNEXPANDED_VISIT_COUNT:
      node.total_visits = 0
      for action, prob in self._evaluator.prior(state):
        node.prior_map[action] = prob
      return self._evaluator.evaluate(state)
    else:
      chosen_action = self.check_expand(
          node, legal_actions)  # add one children at a time?
      if chosen_action != pyspiel.INVALID_ACTION:
        # check if all actions have been expanded, if not, select one?
        # if yes, ucb?
        self.expand_if_necessary(node, chosen_action)
      else:
        chosen_action = self.select_action_tree_policy(node, legal_actions)

      assert chosen_action != pyspiel.INVALID_ACTION

      node.total_visits += 1
      node.child_info[chosen_action].visits += 1
      state.apply_action(chosen_action)
      returns = self.run_simulation(state)
      node.child_info[chosen_action].return_sum += returns[cur_player]
      return returns



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.egt.alpharank_visualizer."""

from absl.testing import absltest

# pylint: disable=g-import-not-at-top
import matplotlib
matplotlib.use("agg")  # switch backend for testing

import mock
import numpy as np

from open_spiel.python.egt import alpharank
from open_spiel.python.egt import alpharank_visualizer
from open_spiel.python.egt import utils
import pyspiel


class AlpharankVisualizerTest(absltest.TestCase):

  @mock.patch("%s.alpharank_visualizer.plt" % __name__)
  def test_plot_pi_vs_alpha(self, mock_plt):
    # Construct game
    game = pyspiel.load_matrix_game("matrix_rps")
    payoff_tables = utils.game_payoffs_array(game)
    _, payoff_tables = utils.is_symmetric_matrix_game(payoff_tables)
    payoffs_are_hpt_format = utils.check_payoffs_are_hpt(payoff_tables)

    # Compute alpharank
    alpha = 1e2
    _, _, pi, num_profiles, num_strats_per_population = (
        alpharank.compute(payoff_tables, alpha=alpha))
    strat_labels = utils.get_strat_profile_labels(payoff_tables,
                                                  payoffs_are_hpt_format)
    num_populations = len(payoff_tables)

    # Construct synthetic pi-vs-alpha history
    pi_list = np.empty((num_profiles, 0))
    alpha_list = []
    for _ in range(2):
      pi_list = np.append(pi_list, np.reshape(pi, (-1, 1)), axis=1)
      alpha_list.append(alpha)

    # Test plotting code (via pyplot mocking to prevent plot pop-up)
    alpharank_visualizer.plot_pi_vs_alpha(
        pi_list.T,
        alpha_list,
        num_populations,
        num_strats_per_population,
        strat_labels,
        num_strats_to_label=0)
    self.assertTrue(mock_plt.show.called)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Implementations of classical fictitious play.

See https://en.wikipedia.org/wiki/Fictitious_play.
"""

import itertools

import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms import exploitability


def _uniform_policy(state):
  legal_actions = state.legal_actions()
  return [(action, 1.0 / len(legal_actions)) for action in legal_actions]


def _callable_tabular_policy(tabular_policy):
  """Turns a tabular policy into a callable.

  Args:
    tabular_policy: A dictionary mapping information state key to a dictionary
      of action probabilities (action -> prob).

  Returns:
    A function `state` -> list of (action, prob)
  """

  def wrap(state):
    infostate_key = state.information_state_string(state.current_player())
    assert infostate_key in tabular_policy
    ap_list = []
    for action in state.legal_actions():
      assert action in tabular_policy[infostate_key]
      ap_list.append((action, tabular_policy[infostate_key][action]))
    return ap_list

  return wrap


class JointPolicy(policy.Policy):
  """A policy for all players in the game."""

  def __init__(self, game, policies):
    """Initializes a joint policy from a table of callables.

    Args:
       game: The game being played.
       policies: A dictionary mapping player number to a function `state` ->
         list of (action, prob).
    """
    super().__init__(game, list(range(game.num_players())))
    self.policies = policies

  def action_probabilities(self, state, player_id=None):
    return dict(self.policies[player_id or state.current_player()](state))


def _full_best_response_policy(br_infoset_dict):
  """Turns a dictionary of best response action selections into a full policy.

  Args:
    br_infoset_dict: A dictionary mapping information state to a best response
      action.

  Returns:
    A function `state` -> list of (action, prob)
  """

  def wrap(state):
    infostate_key = state.information_state_string(state.current_player())
    br_action = br_infoset_dict[infostate_key]
    ap_list = []
    for action in state.legal_actions():
      ap_list.append((action, 1.0 if action == br_action else 0.0))
    return ap_list

  return wrap


def _policy_dict_at_state(callable_policy, state):
  """Turns a policy function into a dictionary at a specific state.

  Args:
    callable_policy: A function from `state` -> lis of (action, prob),
    state: the specific state to extract the policy from.

  Returns:
    A dictionary of action -> prob at this state.
  """

  infostate_policy_list = callable_policy(state)
  infostate_policy = {}
  for ap in infostate_policy_list:
    infostate_policy[ap[0]] = ap[1]
  return infostate_policy


class XFPSolver(object):
  """An implementation of extensive-form fictitious play (XFP).

  XFP is Algorithm 1 in (Heinrich, Lanctot, and Silver, 2015, "Fictitious
  Self-Play in Extensive-Form Games"). Refer to the paper for details:
  http://mlanctot.info/files/papers/icml15-fsp.pdf.
  """

  def __init__(self, game, save_oracles=False):
    """Initialize the XFP solver.

    Arguments:
      game: the open_spiel game object.
      save_oracles: a boolean, indicating whether or not to save all the BR
        policies along the way (including the initial uniform policy). This
        could take up some space, and is only used when generating the meta-game
        for analysis.
    """

    self._game = game
    self._num_players = self._game.num_players()

    # A set of callables that take in a state and return a list of
    # (action, probability) tuples.
    self._oracles = [] if save_oracles else None

    # A set of callables that take in a state and return a list of
    # (action, probability) tuples.
    self._policies = []
    for _ in range(self._num_players):
      self._policies.append(_uniform_policy)
      if save_oracles:
        self._oracles.append([_uniform_policy])

    self._best_responses = [None] * self._num_players
    self._iterations = 0
    self._delta_tolerance = 1e-10
    self._average_policy_tables = []

  def average_policy_tables(self):
    """Returns a dictionary of information state -> dict of action -> prob.

    This is a joint policy (policy for all players).
    """
    return self._average_policy_tables

  def average_policy(self):
    """Returns the current average joint policy (policy for all players)."""
    return JointPolicy(self._game, self._policies)

  def iteration(self):
    self._iterations += 1
    self.compute_best_responses()
    self.update_average_policies()

  def compute_best_responses(self):
    """Updates self._oracles to hold best responses for each player."""
    for i in range(self._num_players):
      # Compute a best response policy to pi_{-i}.
      # First, construct pi_{-i}.
      joint_policy = self.average_policy()
      br_info = exploitability.best_response(self._game,
                                             joint_policy.to_tabular(), i)
      full_br_policy = _full_best_response_policy(
          br_info["best_response_action"])
      self._best_responses[i] = full_br_policy
      if self._oracles is not None:
        self._oracles[i].append(full_br_policy)

  def update_average_policies(self):
    """Update the average policies given the newly computed best response."""

    br_reach_probs = np.ones(self._num_players)
    avg_reach_probs = np.ones(self._num_players)
    self._average_policy_tables = [{} for _ in range(self._num_players)]
    self._recursively_update_average_policies(self._game.new_initial_state(),
                                              avg_reach_probs, br_reach_probs)
    for i in range(self._num_players):
      self._policies[i] = _callable_tabular_policy(
          self._average_policy_tables[i])

  def _recursively_update_average_policies(self, state, avg_reach_probs,
                                           br_reach_probs):
    """Recursive implementation of the average strategy update."""

    if state.is_terminal():
      return
    elif state.is_chance_node():
      for action, _ in state.chance_outcomes():
        new_state = state.clone()
        new_state.apply_action(action)
        self._recursively_update_average_policies(new_state, avg_reach_probs,
                                                  br_reach_probs)
    else:
      player = state.current_player()
      avg_policy = _policy_dict_at_state(self._policies[player], state)
      br_policy = _policy_dict_at_state(self._best_responses[player], state)
      legal_actions = state.legal_actions()
      infostate_key = state.information_state_string(player)
      # First traverse the subtrees.
      for action in legal_actions:
        assert action in br_policy
        assert action in avg_policy
        new_state = state.clone()
        new_state.apply_action(action)
        new_avg_reach = np.copy(avg_reach_probs)
        new_avg_reach[player] *= avg_policy[action]
        new_br_reach = np.copy(br_reach_probs)
        new_br_reach[player] *= br_policy[action]
        self._recursively_update_average_policies(new_state, new_avg_reach,
                                                  new_br_reach)
      # Now, do the updates.
      if infostate_key not in self._average_policy_tables[player]:
        alpha = 1 / (self._iterations + 1)
        self._average_policy_tables[player][infostate_key] = {}
        pr_sum = 0.0
        for action in legal_actions:
          pr = (
              avg_policy[action] + (alpha * br_reach_probs[player] *
                                    (br_policy[action] - avg_policy[action])) /
              ((1.0 - alpha) * avg_reach_probs[player] +
               alpha * br_reach_probs[player]))
          self._average_policy_tables[player][infostate_key][action] = pr
          pr_sum += pr
        assert (1.0 - self._delta_tolerance <= pr_sum <=
                1.0 + self._delta_tolerance)

  def sample_episode(self, state, policies):
    """Samples an episode according to the policies, starting from state.

    Args:
      state: Pyspiel state representing the current state.
      policies: List of policy representing the policy executed by each player.

    Returns:
      The result of the call to returns() of the final state in the episode.
        Meant to be a win/loss integer.
    """

    if state.is_terminal():
      return np.array(state.returns(), dtype=np.float32)
    elif state.is_chance_node():
      outcomes = []
      probs = []
      for action, prob in state.chance_outcomes():
        outcomes.append(action)
        probs.append(prob)
      outcome = np.random.choice(outcomes, p=probs)
      state.apply_action(outcome)
      return self.sample_episode(state, policies)
    else:
      player = state.current_player()
      state_policy = _policy_dict_at_state(policies[player], state)
      actions = []
      probs = []
      for action in state_policy:
        actions.append(action)
        probs.append(state_policy[action])
      action = np.random.choice(actions, p=probs)
      state.apply_action(action)
      return self.sample_episode(state, policies)

  def sample_episodes(self, policies, num):
    """Samples episodes and averages their returns.

    Args:
      policies: A list of policies representing the policies executed by each
        player.
      num: Number of episodes to execute to estimate average return of policies.

    Returns:
      Average episode return over num episodes.
    """

    totals = np.zeros(self._num_players)
    for _ in range(num):
      totals += self.sample_episode(self._game.new_initial_state(), policies)
    return totals / num

  def get_empirical_metagame(self, sims_per_entry, seed=None):
    """Gets a meta-game tensor of utilities from episode samples.

    The tensor is a cross-table of all the saved oracles and initial uniform
    policy.

    Args:
        sims_per_entry: number of simulations (episodes) to perform per entry in
          the tables, i.e. each is a crude Monte Carlo estimate
        seed: the seed to set for random sampling, for reproducibility

    Returns:
        the K^n (KxKx...K, with dimension n) meta-game tensor where n is the
        number of players and K is the number of strategies (one more than the
        number of iterations of fictitious play since the initial uniform
        policy is included).
    """

    if seed is not None:
      np.random.seed(seed=seed)
    assert self._oracles is not None
    num_strategies = len(self._oracles[0])
    # Each metagame will be (num_strategies)^self._num_players.
    # There are self._num_player metagames, one per player.
    meta_games = []
    for _ in range(self._num_players):
      shape = [num_strategies] * self._num_players
      meta_game = np.ndarray(shape=shape, dtype=np.float32)
      meta_games.append(meta_game)
    for coord in itertools.product(
        range(num_strategies), repeat=self._num_players):
      policies = []
      for i in range(self._num_players):
        iteration = coord[i]
        policies.append(self._oracles[i][iteration])
      utility_estimates = self.sample_episodes(policies, sims_per_entry)
      for i in range(self._num_players):
        meta_games[i][coord] = utility_estimates[i]
    return meta_games



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python spiel example."""

import logging
import sys
from absl import app
from absl import flags
import numpy as np

import pyspiel
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import random_agent
from open_spiel.python.algorithms import tabular_qlearner

from open_spiel.python import games  # pylint: disable=unused-import
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_integer("num_episodes", int(100), "Number of train episodes.")
flags.DEFINE_boolean(
    "interactive_play",
    True,
    "Whether to run an interactive play with the agent after training.",
)

"""
_DICT_CHESS_CELL = {6: "雷", 5: "师", 4: "旅", 3: "团", 2: "炸", 1: "棋", 0: "空"}
def pretty_board(time_step):
    info_state = time_step.observations["info_state"]
    info_board = np.reshape(info_state[0], (13,6,3))
    board = [[f"\033[;;m空\033[0m"]*3]*6
    for chess_axis in range(1, len(info_board)):
        for i in range(6):
            for j in range(3):
                if info_board[chess_axis][i][j]:
                    if chess_axis >= 6:
                        axis = chess_axis - 6
                        board[i][j] = f"\033[;30;42m{_DICT_CHESS_CELL[axis]}\033[0m"
                    else:
                        axis = chess_axis + 0
                        board[i][j] = f"\033[;30;43m{_DICT_CHESS_CELL[axis]}\033[0m"
    string_board = "\n".join("".join([str(chess) for chess in row]) for row in board)
    return string_board
"""


def command_line_action(time_step):
    """Gets a valid action from the user on the command line."""
    current_player = time_step.observations["current_player"]
    legal_actions = time_step.observations["legal_actions"][current_player]
    action = -1
    while action not in legal_actions:
        print("Choose an action from {}:".format(legal_actions))
        sys.stdout.flush()
        action_str = input()
        try:
            action = int(action_str)
        except ValueError:
            continue
    return action


def eval_against_random_bots(env, trained_agents, random_agents, num_episodes):
    """Evaluates `trained_agents` against `random_agents` for `num_episodes`."""
    wins = np.zeros(2)
    for player_pos in range(2):
        if player_pos == 0:
            cur_agents = [trained_agents[0], random_agents[1]]
        else:
            cur_agents = [random_agents[0], trained_agents[1]]
        for _ in range(num_episodes):
            time_step = env.reset()
            while not time_step.last():
                player_id = time_step.observations["current_player"]
                agent_output = cur_agents[player_id].step(time_step, is_evaluation=True)
                time_step = env.step([agent_output.action])
            if time_step.rewards[player_pos] > 0:
                wins[player_pos] += 1
    return wins / num_episodes


class JunQiEnv(rl_environment.Environment):
    def get_stringed_board(self):
        return self._state.serialize()

    def get_stringed_action(self):
        actions = self._state.legal_actions()
        return "\n".join(self._state.serialize_action(action) for action in actions)


def main(_):
    num_players = 2
    game = "junqi"

    env = JunQiEnv(game)
    num_actions = env.action_spec()["num_actions"]

    agents = [
        tabular_qlearner.QLearner(player_id=idx, num_actions=num_actions)
        for idx in range(num_players)
    ]

    # random agents for evaluation
    random_agents = [
        random_agent.RandomAgent(player_id=idx, num_actions=num_actions)
        for idx in range(num_players)
    ]

    # 1. Train the agents
    training_episodes = FLAGS.num_episodes
    for cur_episode in range(training_episodes):
        #if cur_episode % int(1e4) == 0:
        win_rates = eval_against_random_bots(env, agents, random_agents, 1000)
        print(f"Starting episode {cur_episode}, win_rates {win_rates}")
        time_step = env.reset()
        while not time_step.last():
            player_id = time_step.observations["current_player"]
            agent_output = agents[player_id].step(time_step)
            time_step = env.step([agent_output.action])

        # Episode is over, step all agents with final info state.
        for agent in agents:
            agent.step(time_step)

    if not FLAGS.interactive_play:
        return

    # 2. Play from the command line against the trained agent.
    human_player = 0
    while True:
        print("You are playing as %s", "O" if human_player else "X")
        time_step = env.reset()
        while not time_step.last():
            player_id = time_step.observations["current_player"]
            if player_id == human_player:
                agent_out = agents[human_player].step(time_step, is_evaluation=True)
                # print("\n%s", agent_out)
                print(env.get_stringed_board(), end="\n\n")
                print(env.get_stringed_action())
                action = command_line_action(time_step)
            else:
                agent_out = agents[1 - human_player].step(time_step, is_evaluation=True)
                action = agent_out.action
            time_step = env.step([action])

        print(env.get_stringed_board(), end="\n\n")

        print("End of game!")
        if time_step.rewards[human_player] > 0:
            print("You win")
        elif time_step.rewards[human_player] < 0:
            print("You lose")
        else:
            print("Draw")
        # Switch order of players
        human_player = 1 - human_player


if __name__ == "__main__":
    app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Joint Policy-Space Response Oracles.

An implementation of JSPRO, described in https://arxiv.org/abs/2106.09435.

Bibtex / Cite:

```
@misc{marris2021multiagent,
    title={Multi-Agent Training beyond Zero-Sum with Correlated Equilibrium
           Meta-Solvers},
    author={Luke Marris and Paul Muller and Marc Lanctot and Karl Tuyls and
            Thore Graepel},
    year={2021},
    eprint={2106.09435},
    archivePrefix={arXiv},
    primaryClass={cs.MA}
}
```
"""

from absl import app
from absl import flags

from open_spiel.python.algorithms import jpsro
import pyspiel


GAMES = (
    "kuhn_poker_2p",
    "kuhn_poker_3p",
    "kuhn_poker_4p",
    "leduc_poker_2p",
    "leduc_poker_3p",
    "leduc_poker_4p",
    "trade_comm_2p_2i",
    "trade_comm_2p_3i",
    "trade_comm_2p_4i",
    "trade_comm_2p_5i",
    "tiny_bridge_2p",
    "tiny_bridge_4p",
    "sheriff_2p_1r",
    "sheriff_2p_2r",
    "sheriff_2p_3r",
    "sheriff_2p_gabriele",
    "goofspiel_2p_3c_total",
    "goofspiel_2p_4c_total",
    "goofspiel_2p_5c_total",
    "goofspiel_2p_5c_total",
    "goofspiel_2p_5c_dsc_total",
    "goofspiel_2p_5c_dsc_pt_diff",
)

FLAGS = flags.FLAGS

# Game.
flags.DEFINE_string(
    "game", "kuhn_poker_3p",
    "Game and settings name.")

# JPSRO - General.
flags.DEFINE_integer(
    "iterations", 40,
    "Number of JPSRO iterations.",
    lower_bound=0)
flags.DEFINE_integer(
    "seed", 1,
    "Pseduo random number generator seed.")
flags.DEFINE_enum(
    "policy_init", "uniform", jpsro.INIT_POLICIES,
    "Initial policy sampling strategy.")
flags.DEFINE_enum(
    "update_players_strategy", "all", jpsro.UPDATE_PLAYERS_STRATEGY,
    "Which player's policies to update at each iteration.")

# JPSRO - Best Response.
flags.DEFINE_enum(
    "target_equilibrium", "cce", jpsro.BRS,
    "The target equilibrium, either ce or cce.")
flags.DEFINE_enum(
    "br_selection", "largest_gap", jpsro.BR_SELECTIONS,
    "The best response operator. Primarily used with CE target equilibrium.")

# JPSRO - Meta-Solver.
flags.DEFINE_enum(
    "train_meta_solver", "mgcce", jpsro.META_SOLVERS,
    "Meta-solver to use for training.")
flags.DEFINE_enum(
    "eval_meta_solver", "mwcce", jpsro.META_SOLVERS,
    "Meta-solver to use for evaluation.")
flags.DEFINE_bool(
    "ignore_repeats", False,
    "Whether to ignore policy repeats when calculating meta distribution. "
    "This is relevant for some meta-solvers (such as Maximum Gini) that will "
    "spread weight over repeats. This may or may not be a desireable property "
    "depending on how one wishes to search the game space. A uniform "
    "meta-solver requires this to be False.")
flags.DEFINE_float(
    "action_value_tolerance", -1.0,
    "If non-negative, use max-entropy best-responses with specified tolerance "
    "on action-value. If negative, the best-response operator will return a "
    "best-response policy that deterministically chooses the first action with "
    "maximum action-value in each state.")


def get_game(game_name):
  """Returns the game."""

  if game_name == "kuhn_poker_2p":
    game_name = "kuhn_poker"
    game_kwargs = {"players": int(2)}
  elif game_name == "kuhn_poker_3p":
    game_name = "kuhn_poker"
    game_kwargs = {"players": int(3)}
  elif game_name == "kuhn_poker_4p":
    game_name = "kuhn_poker"
    game_kwargs = {"players": int(4)}

  elif game_name == "leduc_poker_2p":
    game_name = "leduc_poker"
    game_kwargs = {"players": int(2)}
  elif game_name == "leduc_poker_3p":
    game_name = "leduc_poker"
    game_kwargs = {"players": int(3)}
  elif game_name == "leduc_poker_4p":
    game_name = "leduc_poker"
    game_kwargs = {"players": int(4)}

  elif game_name == "trade_comm_2p_2i":
    game_name = "trade_comm"
    game_kwargs = {"num_items": int(2)}
  elif game_name == "trade_comm_2p_3i":
    game_name = "trade_comm"
    game_kwargs = {"num_items": int(3)}
  elif game_name == "trade_comm_2p_4i":
    game_name = "trade_comm"
    game_kwargs = {"num_items": int(4)}
  elif game_name == "trade_comm_2p_5i":
    game_name = "trade_comm"
    game_kwargs = {"num_items": int(5)}

  elif game_name == "tiny_bridge_2p":
    game_name = "tiny_bridge_2p"
    game_kwargs = {}
  elif game_name == "tiny_bridge_4p":
    game_name = "tiny_bridge_4p"
    game_kwargs = {}  # Too big game.

  elif game_name == "sheriff_2p_1r":
    game_name = "sheriff"
    game_kwargs = {"num_rounds": int(1)}
  elif game_name == "sheriff_2p_2r":
    game_name = "sheriff"
    game_kwargs = {"num_rounds": int(2)}
  elif game_name == "sheriff_2p_3r":
    game_name = "sheriff"
    game_kwargs = {"num_rounds": int(3)}
  elif game_name == "sheriff_2p_gabriele":
    game_name = "sheriff"
    game_kwargs = {
        "item_penalty": float(1.0),
        "item_value": float(5.0),
        "max_bribe": int(2),
        "max_items": int(10),
        "num_rounds": int(2),
        "sheriff_penalty": float(1.0),
    }

  elif game_name == "goofspiel_2p_3c_total":
    game_name = "goofspiel"
    game_kwargs = {
        "players": int(2),
        "returns_type": "total_points",
        "num_cards": int(3)}
  elif game_name == "goofspiel_2p_4c_total":
    game_name = "goofspiel"
    game_kwargs = {
        "players": int(2),
        "returns_type": "total_points",
        "num_cards": int(4)}
  elif game_name == "goofspiel_2p_5c_total":
    game_name = "goofspiel"
    game_kwargs = {
        "imp_info": True,
        "egocentric": True,
        "players": int(2),
        "returns_type": "total_points",
        "num_cards": int(5)
    }
  elif game_name == "goofspiel_2p_5c_dsc_total":
    game_name = "goofspiel"
    game_kwargs = {
        "imp_info": True,
        "egocentric": True,
        "points_order": "descending",
        "players": int(2),
        "returns_type": "total_points",
        "num_cards": int(5)
    }
  elif game_name == "goofspiel_2p_5c_dsc_pt_diff":
    game_name = "goofspiel"
    game_kwargs = {
        "imp_info": True,
        "egocentric": True,
        "points_order": "descending",
        "players": int(2),
        "returns_type": "point_difference",
        "num_cards": int(5)
    }

  else:
    raise ValueError("Unrecognised game: %s" % game_name)

  return pyspiel.load_game_as_turn_based(game_name, game_kwargs)


def main(argv):
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")
  game = get_game(FLAGS.game)
  jpsro.run_loop(
      game=game,
      game_name=FLAGS.game,
      seed=FLAGS.seed,
      iterations=FLAGS.iterations,
      policy_init=FLAGS.policy_init,
      update_players_strategy=FLAGS.update_players_strategy,
      target_equilibrium=FLAGS.target_equilibrium,
      br_selection=FLAGS.br_selection,
      train_meta_solver=FLAGS.train_meta_solver,
      eval_meta_solver=FLAGS.eval_meta_solver,
      action_value_tolerance=FLAGS.action_value_tolerance,
      ignore_repeats=FLAGS.ignore_repeats)


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Benchmark performance of games by counting the number of rollouts."""

import random
import time

from absl import app
from absl import flags
from absl import logging
import pandas as pd

from open_spiel.python import games  # pylint: disable=unused-import
from open_spiel.python.mfg import games as mfg_games  # pylint: disable=unused-import
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_string(
    "games", "*", "Benchmark only specific games (semicolon separated). "
    "Use * to benchmark all (loadable) games.")
flags.DEFINE_float("time_limit", 10., "Time limit per game (in seconds).")
flags.DEFINE_integer("give_up_after", 100,
                     "Give up rollout when the history length is exceeded.")
flags.DEFINE_bool(
    "if_simultaneous_convert_to_turn_based", False,
    "If True, load any simultaneous game as turn based for the benchmark.")


def _rollout_until_timeout(game_name,
                           time_limit,
                           give_up_after,
                           if_simultaneous_convert_to_turn_based=False):
  """Run rollouts on the specified game until the time limit.

  Args:
    game_name:      str
    time_limit:     In number of seconds
    give_up_after:  Cuts off trajectories longer than specified
    if_simultaneous_convert_to_turn_based: if the game is simultaneous and this
      boolean is true, then the game is loaded as a turn based game.

  Returns:
    A dict of collected statistics.
  """
  game = pyspiel.load_game(game_name)
  if game.get_type().dynamics == pyspiel.GameType.Dynamics.MEAN_FIELD:
    raise NotImplementedError(
        "Benchmark on mean field games is not available yet.")
  if (game.get_type().dynamics == pyspiel.GameType.Dynamics.SIMULTANEOUS and
      if_simultaneous_convert_to_turn_based):
    game = pyspiel.convert_to_turn_based(game)
  is_time_out = lambda t: time.time() - t > time_limit
  num_rollouts = 0
  num_giveups = 0
  num_moves = 0
  start = time.time()
  while not is_time_out(start):
    state = game.new_initial_state()
    while not state.is_terminal():
      if len(state.history()) > give_up_after:
        num_giveups += 1
        break
      if state.is_simultaneous_node():

        def random_choice(actions):
          if actions:
            return random.choice(actions)
          return 0

        actions = [
            random_choice(state.legal_actions(i))
            for i in range(state.num_players())
        ]
        state.apply_actions(actions)
      else:
        action = random.choice(state.legal_actions(state.current_player()))
        state.apply_action(action)
      num_moves += 1
    num_rollouts += 1
  time_elapsed = time.time() - start
  return dict(
      game_name=game_name,
      ms_per_rollouts=time_elapsed / num_rollouts * 1000,
      ms_per_moves=time_elapsed / num_moves * 1000,
      giveups_per_rollout=num_giveups / num_rollouts,
      time_elapsed=time_elapsed
  )


def main(_):
  if FLAGS.games == "*":
    games_list = [
        game.short_name
        for game in pyspiel.registered_games()
        if game.default_loadable
    ]
  else:
    games_list = FLAGS.games.split(";")

  logging.info("Running benchmark for %s games.", len(games_list))
  logging.info("This will take approximately %d seconds.",
               len(games_list) * FLAGS.time_limit)

  game_stats = []
  for game_name in games_list:
    logging.info("Running benchmark on %s", game_name)
    game_stats.append(
        _rollout_until_timeout(game_name, FLAGS.time_limit, FLAGS.give_up_after,
                               FLAGS.if_simultaneous_convert_to_turn_based))

  with pd.option_context("display.max_rows", None,
                         "display.max_columns", None,
                         "display.width", 200):
    df = pd.DataFrame(game_stats)
    # Use nice header names.
    df.rename(columns={
        "game_name": "Game",
        "ms_per_rollouts": "msec/rollout",
        "ms_per_moves": "msec/move",
        "giveups_per_rollout": "Give ups/rollouts",
        "time_elapsed": "Time elapsed [sec]"
    }, inplace=True)

    print("---")
    print("Results for following benchmark configuration:")
    print("time_limit =", FLAGS.time_limit)
    print("give_up_after =", FLAGS.give_up_after)
    print("---")
    print(df)


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.games.data."""

from absl.testing import absltest
from absl.testing import parameterized
from open_spiel.python.algorithms import exploitability
from open_spiel.python.games import data
import pyspiel


class NashEquilibriumtest(parameterized.TestCase):

  @parameterized.parameters((0.), (0.1), (1 / 3))
  def test_exploitability_is_zero_on_nash(self, alpha):
    # A similar test exists in:
    # open_spiel/python/algorithms/exploitability_test.py
    game = pyspiel.load_game("kuhn_poker")
    policy = data.kuhn_nash_equilibrium(alpha=alpha)
    expl = exploitability.exploitability(game, policy)
    self.assertAlmostEqual(0, expl)


if __name__ == "__main__":
  absltest.main()
